#include "page_queue.h"

/* Keep waiting for a potential queue non-empty condition.
   Then, attempt to pop an entry from the queue. */

void PageQueue::pop(PageOwner **owner, Flexpage **flexpage)
{
    std::unique_lock<std::mutex> guard(_lock);
    QueueEntry entry;

    while (1)
    {
        if (_pop(&entry))
        {
            *owner = entry.owner;
            *flexpage = entry.flexpage;
            return;
        }
        else
            _counter.wait(guard);
    }
}

/* Check the queue for entries, returning false if no entries are available,
   returning true and providing the details if an entry can be removed from the
   front of the queue. */

bool PageQueue::_pop(QueueEntry *entry)
{
    if (_queue.empty())
        return false;

    *entry = _queue.front();
    _queue.pop_front();

    /* Remove any position reference for the flexpage. */

    Positions::iterator position = _positions.find(entry->flexpage);

    if (position != _positions.end())
        _positions.erase(position);

    return true;
}

/* Push an entry for the given owner and flexpage to the queue. */

void PageQueue::push(PageOwner *owner, Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Record the entry and a position reference for the flexpage. */

    _queue.push_back((QueueEntry) {flexpage, owner});

    Queue::iterator last = _queue.end();
    last--;
    _positions.insert(Position(flexpage, last));

    _counter.notify_one();
}

/* Push an entry to the front of the queue. */

void PageQueue::push_front(PageOwner *owner, Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    _queue.push_back((QueueEntry) {flexpage, owner});
    _positions.insert(Position(flexpage, _queue.begin()));

    _counter.notify_one();
}

/* Remove an entry for the given owner and flexpage from the queue. */

bool PageQueue::remove(PageOwner *owner, Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    Positions::iterator position = _positions.find(flexpage);

    if (position == _positions.end())
        return false;

    /* The found owner may be different from the requesting owner or even NULL
       if another owner has acquired and then purged its pages. Such a purged
       flexpage is not immediately usable, however. */

    Queue::iterator entry = position->second;

    if ((entry->owner == NULL) || (entry->owner != owner))
        return false;

    _queue.erase(entry);
    _positions.erase(position);

    return true;
}

// vim: tabstop=4 expandtab shiftwidth=4
